#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include <iiimp-data.h>
#include <iiimp-opcode.h>

#include "iiimp-dataP.h"


IIIMP_message *
iiimp_read_ns_new(
    IIIMP_data_s *	data_s,
    IIIMP_card16        im_id,
    IIIMP_card16        ic_id,
    IIIMP_card32        listener_id,
    IIIMP_string *	filename)
{
    IIIMP_message *	m;

    m = (IIIMP_message *)malloc(sizeof (IIIMP_message));
    if (NULL == m) {
	data_s->status = IIIMP_DATA_MALLOC_ERROR;
	return NULL;
    }
    m->opcode = IM_READ_NS;
    m->im_id = im_id;
    m->ic_id = ic_id;
    m->v.im_read_ns.listener_id = listener_id;
    m->v.im_read_ns.filename = filename;
    return m;
}

void
iiimp_read_ns_delete(IIIMP_data_s * data_s, IIIMP_message * m)
{
    if (NULL == m) return;
    iiimp_string_delete(data_s, m->v.im_read_ns.filename);
    free(m);
    return;
}

uchar_t *
iiimp_read_ns_pack(
    IIIMP_data_s *	data_s,
    IIIMP_card16        im_id,
    IIIMP_card16        ic_id,
    IIIMP_card32        listener_id,
    IIIMP_string *      filename,
    size_t *		buf_size)
{
    size_t		nbyte;
    int			length;
    uchar_t *		buf;
    size_t		rest;
    uchar_t *		p;

    nbyte = 0
	+ 2                 /* input method id */
	+ 2                 /* input context id */
	+ 4                 /* NSID */
	+ filename->nbyte;  /* Filename */

    length = (nbyte >> 2);
    *buf_size = (1 + 3 + nbyte);

    buf = (uchar_t *)malloc(1 + 3 + nbyte);
    if (NULL == buf) {
	data_s->status = IIIMP_DATA_MALLOC_ERROR;
	return NULL;
    }

    PUT_PACKET_HEADER(buf, IM_READ_NS, length);

    rest = nbyte;
    p = (buf + 4);

    PUTU16(im_id, rest, p, data_s->byte_swap);
    PUTU16(ic_id, rest, p, data_s->byte_swap);
    PUTU32(listener_id, rest, p, data_s->byte_swap);
    iiimp_string_pack(data_s, filename, &rest, &p);

    return buf;
}

IIIMP_message *
iiimp_read_ns_unpack(
    IIIMP_data_s *      data_s,
    IIIMP_card7         opcode,
    size_t *            nbyte,
    const uchar_t **    ptr)
{
    IIIMP_message *        m;
    IIIMP_im_read_ns_v *   v;
    size_t                 rest;
    const uchar_t *        p;
    int                    len;

    rest = *nbyte;
    p = *ptr;

    if (rest < (2 + 2 + 4 + 4)) {
        data_s->status = IIIMP_DATA_INVALID;
        return NULL;
    }

    m = (IIIMP_message *)malloc(sizeof (IIIMP_message));
    if (NULL == m) {
        data_s->status = IIIMP_DATA_MALLOC_ERROR;
        return NULL;
    }
    v = &(m->v.im_read_ns);

    m->opcode = opcode;

    GETU16(m->im_id, rest, p, data_s->byte_swap);
    GETU16(m->ic_id, rest, p, data_s->byte_swap);
    GETU32(v->listener_id, rest, p, data_s->byte_swap);
    v->filename = iiimp_string_unpack(data_s, &rest, &p, rest);

    return m;
}

void
iiimp_read_ns_print(
    IIIMP_data_s *      data_s,
    IIIMP_message *     m)
{
    iiimp_message_header_print(data_s, m->opcode, m->im_id, m->ic_id);
    (void)fprintf(data_s->print_fp, "\tlistener_id = %d\n", m->v.im_read_ns.listener_id);
    (void)fprintf(data_s->print_fp, "\tfilename = ");
    iiimp_string_print(data_s, m->v.im_read_ns.filename);
    (void)fputc('\n', data_s->print_fp);
}

/* Local Variables: */
/* c-file-style: "iiim-project" */
/* End: */
